# Copyright (c) 2021 Matteo Redaelli
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import qsAPI


def update_custom_property_with_users_list(
    qrs, custom_property_name, user_directory, dryrun=True, threshold=100
):
    """Update a custom property (usually UserAccess) with the list of Qliksense users. There is a threshold,
    just to be sure that at least a fixes nunber of user should be found before updating teh custom property
    """
    result = qrs.driver.get(
        "/qrs/custompropertydefinition/full",
        {"filter": "name eq '{name}'".format(name=custom_property_name)},
    ).json()
    assert len(result) == 1
    user_access = result[0]
    logging.debug(
        "Found custom property with id="
        + user_access["id"]
        + " and name="
        + user_access["name"]
    )
    logging.info("extracting all qliksense users...")
    users = qrs.UserGet(
        pFilter="userDirectory eq '{directory}' and removedExternally eq False".format(
            directory=user_directory
        )
    )
    logging.info("Found {tot} users".format(tot=len(users)))
    if len(users) < check_count:
        logging.warning(
            "The custom poperty will not be updated: too few users, the threshold is {threshold}".format(
                threshold=threshold
            )
        )
        return
    names = list(map(lambda u: u["userId"], users))

    user_access["choiceValues"] = names
    if not dryrun:
        return qrs.driver.put(
            "/qrs/CustomPropertyDefinition/{id}".format(id=user_access["id"]),
            data=user_access,
        )



#!/usr/bin/env python3
# Copyright (c) 2021 Matteo Redaelli
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Command line
"""
import json
import logging
import fire
import qsAPI
from pyqlikengine.engine import QixEngine
import qsense


def connect_qix_engine(host, certfile, keyfile, ca_certs, user_directory, user_id):
    """connect to the qix engine"""
    qixe = QixEngine(
        url=host,
        user_directory=user_directory,
        user_id=user_id,
        ca_certs=ca_certs,
        certfile=certfile,
        keyfile=keyfile,
    )
    return qixe


class Qsense:
    """qsense is a python library and command line tool for Qliksense administrators"""

    def get(self, host, certificate, path, service="qrs", port=4242):
        """generic get http from Qlik (service can be qrs or qps)"""
        if service.upper() == "QPS":
            qrs = qsAPI.QPS(proxy=host, certificate=certificate, port=port)
        else:
            qrs = qsAPI.QRS(proxy=host, certificate=certificate, port=port)

        resp = qrs.driver.get(path)
        print(resp)
        if resp.ok:
            return json.dumps(resp.json())
        else:
            return resp

    def user_sessions(
        self, host, certificate, userdirectory, userid, virtualproxy="", port=4243
    ):
        """user sessions"""
        qps = qsAPI.QPS(proxy=host, certificate=certificate, port=port)

        ##        for vp in virtualproxy:
        vp = virtualproxy
        print("Virtual proxy '{vp}'".format(vp=vp))
        path = "/qps/{vp}/user/{userdirectory}/{userid}".format(
            vp=vp, userdirectory=userdirectory, userid=userid
        )
        resp = qps.driver.get(path)
        # print(resp)
        if resp.ok:
            return json.dumps(resp.json())
        else:
            return resp

    def users_with_unpublished_apps(
        self,
        host,
        certificate,
        threshold=100,
    ):
        """Find users with too many unpublished apps"""
        qrs = qsAPI.QRS(proxy=host, certificate=certificate)
        filter = "published eq False"
        ## find unpublished apps
        apps = qrs.driver.get(
            "/qrs/app/full",
            {"filter": filter},
        ).json()
        logging.debug("Found %d apps." % len(apps))
        ## extract owner IDs from apps
        users = list(map(lambda a: a["owner"]["name"] + "|" + a["owner"]["id"], apps))

        ## frequency
        users_freq = qsense.utils.count_frequency(users)
        ## filter highest values
        threshold = int(threshold)

        for u, count in users_freq.items():
            u_name = u.split("|")[0]
            u_id = u.split("|")[1]
            if count > threshold:
                logging.info("%s: %d" % (u, count))

    def add_entity(self, host, certificate, entity, filename, newid=False):
        """add a new entity (user, stream, dataconnection,...)"""
        qrs = qsAPI.QRS(proxy=host, certificate=certificate)

        with open(filename) as f:
            bodies = json.load(f)
        if not bodies:
            logging.error("Cannot read body from filename. Bye")
            return
        if not isinstance(bodies, list):
            logging.error("A list json object was expected. Bye")
            return
        for body in bodies:
            if newid:
                if "id" in body.keys():
                    logging.debug(
                        "Removing the ID {id} from the entity".format(id=body["id"])
                    )
                    del body["id"]
            logging.debug("Adding entity {entity}".format(entity=entity))
            logging.debug(body)
            result = qrs.driver.post("/qrs/{entity}".format(entity=entity), data=body)
            logging.debug(result)
            # logging.debug(result.json())

    def update_entity(self, host, certificate, entity, filename):
        """update an entity (user, stream, dataconnection,...)"""
        qrs = qsAPI.QRS(proxy=host, certificate=certificate)

        with open(filename) as f:
            bodies = json.load(f)
        if not bodies:
            logging.error("Cannot read body from filename. Bye")
            return
        if not isinstance(bodies, list):
            logging.error("A list json object was expected. Bye")
            return
        for body in bodies:
            id = body["id"]
            logging.debug(
                "Updating resource {entity} with id {id}".format(entity=entity, id=id)
            )
            logging.debug(body)
            result = qrs.driver.put(
                "/qrs/{entity}/{id}".format(entity=entity, id=id), data=body
            )
            logging.debug(result)
            logging.debug(result.json())

    def entity(self, host, certificate, entity, id="full", filter=None):
        """Get a specific entity by ID or entity list or count"""
        qrs = qsAPI.QRS(proxy=host, certificate=certificate)
        result = qrs.driver.get(
            "/qrs/{entity}/{id}".format(entity=entity, id=id),
            {"filter": filter},
        ).json()
        print(json.dumps(result))

    def healthcheck(self, host, certificate, port=4747):
        """Get a specific entity by ID or entity list or count"""
        qps = qsAPI.QPS(proxy=host, certificate=certificate, port=port)
        result = qps.driver.get("/healthcheck")
        print(json.dumps(result.json()))

    def deallocate_unused_analyzer_licenses(self, host, certificate, days, dryrun=True):
        """Deallocate analyzer license not used for N days"""
        qrs = qsAPI.QRS(proxy=host, certificate=certificate)
        qsense.license.deallocate_unused_analyzer_licenses(qrs, days, dryrun)

    def deallocate_analyzer_licenses_for_professionals(
        self, host, certificate, dryrun=True
    ):
        """Deallocate analyzer license fom users with a professional license"""
        qrs = qsAPI.QRS(proxy=host, certificate=certificate)
        qsense.license.deallocate_analyzer_licenses_for_professionals(qrs, dryrun)

    def delete_removed_exernally_users(
        self, host, certificate, user_directory, dryrun=True
    ):
        """Delete users that were removed externally (from active directory?)"""
        qrs = qsAPI.QRS(proxy=host, certificate=certificate)
        qsense.users.delete_removed_exernally_users(qrs, user_directory, dryrun)

    def export_apps(
        self, host, certificate, target_path, filter="stream.name ne 'None'"
    ):
        """Export (published or passing any other filter) applications to qvd files"""
        qrs = qsAPI.QRS(proxy=host, certificate=certificate)
        qsense.apps.export_by_filter(qrs, target_path=target_path, pFilter=filter)

    def users(self, host, certificate):
        """Get users with groups"""
        qrs = qsAPI.QRS(proxy=host, certificate=certificate)
        qsense.users.get_users_and_groups(qrs)

    def find_old_apps(
        self,
        host,
        certificate,
        modified_days=1000,
        last_reload_days=1000,
        published=False,
        target_path=".",
        save_meta=True,
        skipdata=True,
        export=False,
        delete=False,
    ):
        """Find old apps using 'modified_date' and 'last_reload_time' filters:
        then you can export them or delete or notify via email the owners"""
        qrs = qsAPI.QRS(proxy=host, certificate=certificate)
        return qsense.apps.find_old_apps(
            qrs,
            modified_days,
            last_reload_days,
            published=published,
            target_path=target_path,
            save_meta=save_meta,
            skipdata=skipdata,
            export=export,
            delete=delete,
        )

    def open_doc(
        self,
        host,
        certfile,
        keyfile,
        ca_certs,
        app_id,
        user_directory="internal",
        user_id="sa_repository",
    ):
        """Load an app in memory, useful for preloading an app / cacha warmer"""
        qixe = connect_qix_engine(
            host, certfile, keyfile, ca_certs, user_directory, user_id
        )
        result = qsense.apps.open_doc(qixe, app_id)
        return result

    def get_app_script(
        self,
        host,
        certfile,
        keyfile,
        ca_certs,
        app_id,
        user_directory="internal",
        user_id="sa_repository",
    ):
        """Extract the ETL script from an app"""
        qixe = connect_qix_engine(
            host, certfile, keyfile, ca_certs, user_directory, user_id
        )
        script = qsense.apps.get_script(qixe, app_id)
        return script

    def get_app_dataconnections(
        self,
        host,
        certfile,
        keyfile,
        ca_certs,
        app_id,
        user_directory="internal",
        user_id="sa_repository",
    ):
        """Extract the dataconnections found in the app script"""
        script = self.app_get_script(
            host, certfile, keyfile, ca_certs, app_id, user_directory, user_id
        )
        return qsense.apps.extract_dataconnections_from_text(script)

    ## TODO / CHECK should return onle teh data connections of the app
    ## but I gget back all data connections

    def get_app_connections(
        self,
        host,
        certfile,
        keyfile,
        ca_certs,
        app_id,
        user_directory="internal",
        user_id="sa_repository",
    ):
        """Extract the connections from an app"""
        qixe = connect_qix_engine(
            host, certfile, keyfile, ca_certs, user_directory, user_id
        )
        result = qsense.apps.get_connections(qixe, app_id)
        return json.dumps(result)

    def update_custom_property_with_users_list(
        self,
        host,
        certificate,
        custom_property_name,
        user_directory,
        dryrun=True,
        threshold=100,
    ):
        """update the values of a custom property with the list of all qliksense users"""
        qrs = qsAPI.QRS(proxy=host, certificate=certificate)
        return qsense.custom_property.update_custom_property_with_users_list(
            qrs, custom_property_name, user_directory, dryrun, threshold
        )


def main():
    """main"""
    logging.basicConfig(level=logging.DEBUG)
    fire.Fire(Qsense)


if __name__ == "__main__":
    main()



# Copyright (c) 2021 Matteo Redaelli
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
import logging
import qsAPI


def extract_mail(u):
    """Extract the mail attribute from the user dictionary"""
    logging.debug("Extract mail from user %s" % str(u))
    if not u:
        logging.error("User is empty")
        return None
    if not "attributes" in u:
        logging.error("User without 'attribute' key")
        return None

    mails = list(filter(lambda a: a["attributeType"] == "Email", u["attributes"]))
    if len(mails) > 0:
        result = mails[0]["attributeValue"]
        logging.debug(result)
        return result
    else:
        logging.error("No mail found for user ")
        return None


def find_mail_from_user_id(qrs, id):
    users = qrs.UserGet(pFilter="id eq %s" % id)
    if len(users) != 1:
        logging.error("Cannot find a user with ID=%" % id)
        mail = None
    else:
        mail = extract_mail(users[0])
    return mail


def delete_removed_exernally_users(qrs, user_directory, dryrun=True):
    users = qrs.UserGet(
        pFilter="userDirectory eq '%s' and removedExternally eq True" % user_directory
    )
    logging.warning("Users to be deleted: %s" % len(users))
    for u in users:
        logging.warning("User to be deleted: %s" % u["name"])
        if not dryrun:
            qrs.UserDelete(u["id"])


def get_users_and_groups(
    qrs,
    pFilter="removedExternally ne True",
    pUserID="full",
    groupFilter="^QLIKSENSE_",
    userAttribute="name",
    sep="\t",
):
    users = qrs.UserGet(pFilter=pFilter, pUserID=pUserID)
    for u in users:
        if "attributes" in u.keys():
            for a in u["attributes"]:
                groupname = a["attributeValue"]
                if re.match(groupFilter, groupname, re.IGNORECASE):
                    print(
                        "{name}{sep}{groupname}\n".format(
                            name=u[userAttribute], groupname=groupname, sep=sep
                        )
                    )


def user_sessions(hosts, certificate, usergroup, userid, vproxies=""):
    for server in hosts.split(","):
        for vproxy in vproxies.split(","):
            qps = qsAPI.QPS(proxy=server, certificate=certificate)
            print(server + " : " + vproxy)
            print(qps.GetUser(usergroup, user).json())



# Copyright (c) 2021 Matteo Redaelli
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# import logging


def count_frequency(my_list):
    """Frequency"""
    count = {}
    for i in my_list:
        count[i] = count.get(i, 0) + 1
    return count



# Copyright (c) 2021 Matteo Redaelli
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="qsense",
    version="0.6.3",
    author="Matteo Redaelli",
    author_email="matteo.redaelli@gmail.com",
    description="qsense is a python library and a command line tool for qliksense",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/matteoredaelli/qsense",
    packages=setuptools.find_packages(),
    license="GPL",
    entry_points={
        "console_scripts": ["qsense=qsense.command_line:main"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=["requests", "requests_ntlm", "qsapi", "pyqlikengine", "fire"],
    python_requires=">=3.6",
)



# Copyright (c) 2021 Matteo Redaelli
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
  license.py

  functions about license
"""

import logging
from datetime import date
from datetime import timedelta
import qsAPI

def deallocate_unused_analyzer_licenses(qrs, days, dryrun=True):
    """ deallocate_unused_analyzer_licenses """
    today = date.today()
    last_used = (today - timedelta(days=days)).strftime("%Y-%m-%d %H:%M:%S")
    logging.debug("Last used date = " + last_used)

    pFilter = "lastUsed lt '{last_used}'".format(last_used=last_used)
    logging.debug("Searching licenses with pFilter= " + str(pFilter))
    analyzers = qrs.driver.get("/qrs/license/analyzeraccesstype/full").json()
    logging.debug("Found %d Analyzer licenses" % len(analyzers))
    for u in analyzers:
        logging.warning("removing analyzer license for user %s" % u["user"]["name"])
        if not dryrun:
            qrs.driver.delete("/qrs/license/analyzeraccesstype/" + u["id"])


def deallocate_analyzer_licenses_for_professionals(qrs, dryrun=True):
    def extract_users(licenses):
        return list(map(lambda u: u["user"]["userId"], licenses))

    def extract_licenses_for_user(licenses, userId):
        return list(filter(lambda u: u["user"]["userId"] == userId, licenses))

    analyzers = qrs.driver.get("/qrs/license/analyzeraccesstype/full").json()
    logging.debug("Analyzer licenses: %d" % len(analyzers))
    analyzers_list = extract_users(analyzers)

    professionals = qrs.driver.get("/qrs/license/professionalaccesstype/full").json()
    logging.debug("Professional licenses: %d" % len(professionals))
    professionals_list = extract_users(professionals)

    to_be_removes = list(set(analyzers_list) & set(professionals_list))
    logging.warning("Licenses to be removed: %d" % len(to_be_removes))

    for u in list(
        map(lambda u: extract_licenses_for_user(analyzers, u), to_be_removes)
    ):
        assert len(u) == 1
        u = u[0]
        logging.warning("removing analyzer license for user %s" % u["user"]["name"])
        if not dryrun:
            qrs.driver.delete("/qrs/license/analyzeraccesstype/" + u["id"])



# Copyright (c) 2021 Matteo Redaelli
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import json
import os
import re
import time
from datetime import date
from datetime import timedelta

# import qsAPI


def extract_dataconnections_from_text(text):
    result = []
    for r in ["LIB CONNECT TO [\['](.*?)[\]']", "[\[']lib://(.*?)[\]']"]:
        reg = re.search(r, text)
        if reg:
            result = result + list(reg.groups())
    return result


def export_app(qrs, target_path, app, save_meta=True, skipdata=True):
    """Export an application to a qvd file and its definition to a json file"""
    logging.info("Exporting app: " + str(app))
    if "stream" in app and "name" in app["name"]:
        filename = app["stream"]["name"]
    else:
        filename = ""
    filename = (
        filename + "-" + app["modifiedByUserName"] + "-" + app["name"] + "-" + app["id"]
    )
    filename = re.sub("[\\\/& .]", "_", filename)
    filename = os.path.join(target_path, filename)
    if save_meta:
        with open(filename + ".json", "w") as f:
            f.write(json.dumps(app, indent=4))
    return qrs.AppExport(app["id"], filename + ".qvd", skipdata=skipdata)


def export_by_filter(
    qrs, target_path, pFilter="stream.name ne 'None'", save_meta=True, skipdata=True
):
    apps = qrs.AppGet(pFilter=pFilter)
    for app in apps:
        export_app(
            qrs,
            target_path=target_path,
            app=app,
            save_meta=save_meta,
            skipdata=skipdata,
        )
    return apps


def open_doc(qixe, qDocId):
    logging.debug(qDocId)
    return qixe.ega.open_doc(qDocId)


def get_doc(qixe, qDocId):
    logging.debug(qDocId)
    qixe.ega.open_doc(qDocId)
    time.sleep(1)
    doc = qixe.ega.get_active_doc()
    logging.debug(doc)
    return doc


def get_connections(qixe, qDocId):
    doc = get_doc(qixe, qDocId)
    if "qReturn" in doc and "qHandle" in doc["qReturn"]:
        h = doc["qReturn"]["qHandle"]
        logging.debug(qixe.eaa.get_all_infos(h))
        return qixe.eaa.get_connections(h)
    else:
        logging.error("not found or something strange appened")
        return None


def get_script(qixe, qDocId):
    doc = get_doc(qixe, qDocId)
    if "qReturn" in doc and "qHandle" in doc["qReturn"]:
        h = doc["qReturn"]["qHandle"]
        logging.debug(qixe.eaa.get_all_infos(h))
        return qixe.eaa.get_script(h)["qScript"]
    else:
        logging.error("not found or something strange appened")
        return None


def get_old_apps(qrs, modified_days, last_reload_days, published=False):
    today = date.today()
    modified_date = (today - timedelta(days=modified_days)).strftime(
        "%Y-%m-%d %H:%M:%S"
    )
    logging.debug("Modified date = " + modified_date)

    last_reload_time = (today - timedelta(days=last_reload_days)).strftime(
        "%Y-%m-%d %H:%M:%S"
    )
    logging.debug("Last reload time = " + last_reload_time)
    pFilter = "published eq {published} and modifiedDate lt '{modified}' and lastReloadTime lt '{reload}'".format(
        published=published, modified=modified_date, reload=last_reload_time
    )
    logging.debug("Search apps with pFilter= " + str(pFilter))
    apps = qrs.AppGet(pFilter=pFilter)
    ##l = map(lambda a: { key:value for key,value in a.items()
    ##                    if key in ['id', 'name',
    ##                               'createdDate',
    ##                               'modifiedDate', 'modifiedByUserName',
    ##                               'published', 'lastReloadTime', 'fileSize']}, apps)
    logging.debug("Found {count} apps:".format(count=len(apps)))
    return apps


def find_old_apps(
    qrs,
    modified_days,
    last_reload_days,
    published,
    target_path,
    save_meta,
    skipdata,
    export,
    delete,
):
    if delete and (modified_days < 60 or last_reload_days < 60):
        logging.error("You want to delete too recent apps. Bye")
        return 1
    apps = get_old_apps(qrs, modified_days, last_reload_days, published)
    for app in apps:
        logging.debug("Found app: " + str(app))
        resp = False
        if export:
            logging.warning("Removing app: " + app["id"])
            resp = export_app(
                qrs,
                app=app,
                target_path=target_path,
                save_meta=save_meta,
                skipdata=skipdata,
            )

        if delete:
            logging.warning("Removing app: " + app["id"])
            # An app can deleted if and only if it was successuffly exported to a file
            if resp and resp.status_code == 200:
                qrs.AppDelete(app["id"])
            else:
                logging.error("Cannot export (and then delete) app: " + app["id"])
